/*
 * Copyright © 2016 Vecly OÜ <info@vecly.com>
 *                  Karel Pärlin
 *                  mart.ruutel
 *                  Silver Kits
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include "m433-test-fw.h"
#include "cc1111.h"
#include "rf.h"
#include "usb.h"

#include <string.h>

__xdata uint8_t usbReadBuffer[BUFFER_SIZE];

// Lenght + Hello World from 433 modem
__xdata uint8_t txTestText [28] = {0x1B, 0x48, 0x65, 0x6C, 0x6C, 0x6F, 0x20, 0x57, 0x6F, 0x72, 0x6C, 0x64, 0x20, 0x66, 0x72, 0x6F, 0x6D, 0x20, 0x34, 0x33, 0x33, 0x20, 0x6D, 0x6F, 0x64, 0x65, 0x6D};

__xdata uint16_t dataLen = 0;   // data part length indicator
__xdata uint8_t cmd = 0;        // command
__xdata uint8_t arg = 0;        // command argument

char continuousRssi = 0;

volatile __xdata int safetyCountdownTimer = 0;
volatile __xdata int delayCountdownTimer = 0;

// Operating mode represents current mode.
// Representation of control protocol command.
// There are differences in data processing and repetition depending from mode.
// It is used because there is no single state machine for mode switching.
volatile __xdata char operatingMode = 0;

void initClock(void) {
    // Switch system clock to crystal oscillator
    CLKCON = (CLKCON & ~CLKCON_OSC_MASK) | (CLKCON_OSC_XTAL);

    while (!(SLEEP & SLEEP_XOSC_STB)) {}

    // Timer tick speed - 187.5 kHz
    // System clock speed - 24 MHz
    CLKCON = ((CLKCON & ~(CLKCON_TICKSPD_MASK | CLKCON_CLKSPD_MASK)) |
      (CLKCON_TICKSPD_1_128 | CLKCON_CLKSPD_1));

    while ((CLKCON & (CLKCON_TICKSPD_MASK|CLKCON_CLKSPD_MASK)) !=
      (CLKCON_TICKSPD_1_128 | CLKCON_CLKSPD_1)
        ) {}
    return;
}

void initLeds(void) {
    // LED as output and off
#ifdef __USBEVAL
    P1DIR |= LED_MASK;
    LED = 0;
#endif
#ifdef __M433LEDBOARD
    P1DIR |= LED_MASK;
    LED_RED     = 0;
    LED_BLUE    = 0;
    LED_WHITE   = 0;
    LED_GREEN   = 0;
#endif
}

void allLedsOff(void) {
#ifdef __USBEVAL
    LED = 0;
#endif
#ifdef __M433LEDBOARD
    LED_RED     = 0;
    LED_BLUE    = 0;
    LED_WHITE   = 0;
    LED_GREEN   = 0;
#endif
}

void allLedsOn(void) {
#ifdef __USBEVAL
    LED = 1;
#endif
#ifdef __M433LEDBOARD
    LED_RED     = 1;
    LED_BLUE    = 1;
    LED_WHITE   = 1;
    LED_GREEN   = 1;
#endif
}

void usbUp(void) {
    P1DIR |= USB_MASK;
    USB_ENABLE = 1;
}

void usbDown(void) {
    USB_ENABLE = 0;
    P1DIR &= ~USB_MASK;
}

void runBootloader(void) {

#if defined(__USBEVAL) || defined(__M433LEDBOARD)
    allLedsOn();
#endif
    // Wait a little bit for a user to notice that bootloader was loaded.
    delayCountdownTimer = BOOTLOADER_LOAD_DELAY;
    while(delayCountdownTimer)
        ;

#if defined(__USBEVAL) || defined(__M433LEDBOARD)
    allLedsOff();
#endif

    if(I2SCLKF0 == 0xf0 && I2SCLKF1 == 0x0d)
        I2SCLKF2 = 0x69;
    // Disable interrupts
    EA = 0;
    IEN0 = IEN1 = IEN2 = 0;
    usbDown();
    SLEEP &= ~SLEEP_USB_EN;

    __asm
    ljmp 0x00
    __endasm;
}

void setupTimer1() {
    // Clear Timer 1 channel 1 and overflow interrupt flag
    // CPU interrupt flag (IRCON) for Timer 1 is cleared automatically by hardware.
    T1CTL &= ~T1CTL_OVFIF;

    // Enable Timer 1 interrupts by setting [IEN1.T1IE=1]
    IEN1 |= IEN1_T1IE;

    // TIMER_TIMEOUT sets the period of the timer timeout in units of
    // approximately 43.7 milliseconds.
    T1CC0H = 23;
    T1CC0L = 0x00;

    // set Timer 1 to modulo mode
    T1CTL = TxCTL_OVFIM | TxCTL_DIV_128 | TxCTL_MODE_MODULO;
}

void timer1Isr() __interrupt T1_VECTOR {
    T1CTL &= ~T1CTL_OVFIF;

    if (safetyCountdownTimer > 0) {
        safetyCountdownTimer--;
    }

    if (delayCountdownTimer > 0) {
        delayCountdownTimer--;
    }
}

void usbPutCommandNotImplemented(void){
    usb_putchar(0x11); // DC1
    usb_putchar(0x01);
    usb_putchar(0xFF); // Status message
    usb_putchar(0x01); // Command not implemented
    usb_flush();
}

void reportFwName(void) {
    __xdata uint8_t ci;
    usb_putchar(0x11); // DC1
    usb_putchar(sizeof(FW_NAME)); // LENGTH
    usb_putchar(0x03); // COMMAND
    for( ci = 0; ci < sizeof(FW_NAME); ci++ )
        usb_putchar(FW_NAME[ci]);
    usb_flush();
}

void reportFwVersion(void) {
    __xdata uint8_t ci;
    usb_putchar(0x11); // DC1
    usb_putchar(sizeof(FW_VERSION)); // LENGTH
    usb_putchar(0x04); // COMMAND
    for( ci = 0; ci < sizeof(FW_VERSION); ci++ )
        usb_putchar(FW_VERSION[ci]);
    usb_flush();
}

void setTxPowerCommand(void) {
    switch (arg) {
        case 0:
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
            rfSetTxPower(arg);
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void reportRssiCommand(void) {
    switch (arg) {
        case 0:
            continuousRssi = 0;
            break;
        case 1:
            usb_putchar(0x11); // DC1
            usb_putchar(1); // LENGTH
            usb_putchar(0x0a); // COMMAND
            usb_putchar(lastRssi);
            usb_flush();
            break;
        case 2:
            continuousRssi = 1;
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void txUnmodulatedCarrierCommand(void) {
    switch (arg) {
    // This handling intentionally differs from other commands.
    case 0:
    case 1:
        rfUnmodulatedContinuousTx(arg);
        break;
    default:
        usbPutCommandNotImplemented();
        break;
    }
}

void rxUnmodulatedCarrierCommand(void) {
    switch (arg) {
        case 0:
            continuousRssi = 0;
            operatingMode = 0;
            safetyCountdownTimer = 0;
            break;
        case 1:
            continuousRssi = 1;
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_TIMER;
            rfIdle();
            rfDefaultRegisters();
            rfInitUnmodulatedRx();
            // There is no need to set it from rfHandler
            // because it shall be set only once.
            rfRx();
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void txAskTextCommand(void) {
    switch (arg) {
        case 0:
            operatingMode = 0;
            safetyCountdownTimer = 0;
            break;
        case 1:
            rfIdle();
            rfDefaultRegisters();
            rfInitAskTx();
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_TIMER;
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void txGfskTextCommand(void) {
    switch (arg) {
        case 0:
            operatingMode = 0;
            safetyCountdownTimer = 0;
            break;
        case 1:
            rfIdle();
            rfDefaultRegisters();
            rfInitGfskTx();
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_TIMER;
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void rxGfskTextCommand(void) {
    switch (arg) {
        case 0:
            operatingMode = 0;
            safetyCountdownTimer = 0;
            break;
        case 1:
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_TIMER;
            rfIdle();
            rfDefaultRegisters();
            rfInitGfskRx();
            // There is no need to set it from rfHandler
            // because it shall be set only once.
            rfRx();
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void testLedsCommand(void) {
    switch (arg) {
        case 0:
            operatingMode = 0;
            safetyCountdownTimer = 0;
            allLedsOff();
            break;
        case 1:
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_GPIO_TIMER;
            allLedsOn();
            break;
        default:
            usbPutCommandNotImplemented();
            usb_flush();
            break;
    }
}

void testSafetyTimerCommand(void) {
    switch (arg) {
        case 0:
            operatingMode = 0;
            safetyCountdownTimer = 0;
            break;
        case 1:
            operatingMode = cmd;
            safetyCountdownTimer = SAFETY_TEST_TIMER;
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void commandHandler(void) {
    switch (cmd) {
        case 0x02:
            runBootloader();
            break;
        case 0x03:
            reportFwName();
            break;
        case 0x04:
            reportFwVersion();
            break;
        case 0x06:
            setTxPowerCommand();
            break;
        case 0x07:
            //There is no need to sanitise input.
            //It is done in rfSetChannel anyway.
            rfSetChannel(arg);
            break;
        case 0x08:
            rfIdle();
            break;
        case 0x0A:
            reportRssiCommand();
            break;
        case 0x10:
            txUnmodulatedCarrierCommand();
            break;
        case 0x11:
            rxUnmodulatedCarrierCommand();
            break;
        case 0x12:
            txAskTextCommand();
            break;
        case 0x13:
            txGfskTextCommand();
            break;
        case 0x14:
            rxGfskTextCommand();
            break;
        case 0x15:
            testLedsCommand();
            break;
        case 0x16:
            testSafetyTimerCommand();
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

void usbHandler(void) {
    char c;

    c = usb_pollchar();
    if (c == USB_READ_AGAIN)
        return;
    c = usb_getchar();

    switch (c) {
        case 0x11:
            dataLen = usb_getchar();
            cmd = usb_getchar();
            if( dataLen == 1 ) {
                arg = usb_getchar();
            }
            usb_flush();
            commandHandler();
            break;
        default:
            usbPutCommandNotImplemented();
            break;
    }
}

#ifdef __M433LEDBOARD
void indicateRssi(void) {
    // Calculate RSSI with factor 74 and switch LED's
    __xdata int rssi_dBm;

    if( lastRssi >= 128 ){
        rssi_dBm = (lastRssi - 256) / 2 - 74;
    } else {
        rssi_dBm = lastRssi / 2 - 74;
    }
    if (rssi_dBm < -70) {
        LED_BLUE    = 0;
        LED_WHITE   = 0;
        LED_GREEN   = 0;
    } else if (rssi_dBm < -60) {
        LED_BLUE    = 1;
        LED_WHITE   = 0;
        LED_GREEN   = 0;
    } else if (rssi_dBm < -50) {
        LED_BLUE    = 1;
        LED_WHITE   = 1;
        LED_GREEN   = 0;
    } else if (rssi_dBm < -40) {
        LED_BLUE    = 1;
        LED_WHITE   = 1;
        LED_GREEN   = 1;
    }
}
#endif

void rfHandler(void) {
    if (usbWriteCounter > 0) {
        __xdata uint8_t i = 0;
        // Send RSSI
        if( continuousRssi ){
            usb_putchar(0x11); // DC1
            usb_putchar(1); // LENGTH
            usb_putchar(0x0a); // COMMAND
            usb_putchar(lastRssi);
#ifdef __M433LEDBOARD
            indicateRssi();
#endif
        }
        // Send data from RX buffer
        while(i < usbWriteCounter) {
            usb_putchar(usbWriteBuffer[i++]);
        }
        usb_flush();
        usbWriteCounter = 0;

    }

    // Safety switch to switch operatingMode 0 after set period.
    // Period shall be activated with setting operating mode.
    // Otherwise mode is switched to 0 immediately.
    if ((operatingMode > 0) && (safetyCountdownTimer == 0)) {
        operatingMode = 0;
        usb_putchar(0x11); // DC1
        usb_putchar(1); // LENGTH
        usb_putchar(0xFF); // COMMAND
        usb_putchar(0x02); // Safety timer was triggered.
        usb_flush();

    }
    switch (operatingMode) {
        case 0x12:
            if (delayCountdownTimer == 0) {
                rfTx(txTestText, 27);
                delayCountdownTimer = PERIODIC_TX_DELAY;
            }
            break;
        case 0x13:
            if (delayCountdownTimer == 0) {
                rfTx(txTestText, 27);
            delayCountdownTimer = PERIODIC_TX_DELAY;
            }
            break;
        case 0:
            allLedsOff();
            rfIdle();
            operatingMode = -1;
            break;
        default:
            break;
    }
}

void main(void) {
    initClock();
    initLeds();
    setupTimer1();
    usb_init();
    rfInit();

    EA = 1;

    usbUp();
    while(1) {
        usbHandler();
        rfHandler();
    }
}
